/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include "json_device_web_handler.h"
#include "ds-assistant.h"
#include "database.h"
#include "storage.h"
#include "exceptions.h"

namespace dSA
{


JSONdeviceWebHandler::JSONdeviceWebHandler(struct mg_connection* connection,
                               const struct mg_request_info* request_info) :
                               JSONWebHandler(connection, request_info) {}

boost::shared_ptr<const Device> JSONdeviceWebHandler::getDeviceFromParam()
{
    std::string dsid = m_parameters["dsid"];
    if (dsid.empty())
    {
        throw _Exception("missing device dSID parameter");
    }

    dsid_t device_dsid = util::str2dsid(dsid);

    boost::shared_ptr<const HWBase> base;

    base = Storage::getInstance()->get(device_dsid);

    if (base == NULL)
    {
        throw _Exception("device " + util::str(device_dsid) + " not found");
    }

    if (base->getDeviceType() != dSDevice)
    {
        throw _Exception(util::str(device_dsid) + " is not a device");
    }

    boost::shared_ptr<const Device> device =
                               boost::dynamic_pointer_cast<const Device>(base);

    return device;
}

std::string JSONdeviceWebHandler::blink()
{
    std::ostringstream result;

    boost::shared_ptr<const Device> device = getDeviceFromParam();

    dS_Assistant *dsa = dS_Assistant::getInstance();

    dsa->dS485()->blinkDevice(device->getDSMdSID(), device->getId());

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdeviceWebHandler::resetStatistic()
{
    std::ostringstream result;

    boost::shared_ptr<const Device> device = getDeviceFromParam();

    dS_Assistant *dsa = dS_Assistant::getInstance();

    uint16_t value;
    dsa->dS485()->getStatisticValue(device->getDSMdSID(), device->getId(), 254, &value);

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdeviceWebHandler::getStatisticValueByIndex()
{
    std::ostringstream result;
    std::string indexs = m_parameters["index"];
    if (indexs.empty())
    {
        throw _Exception("Parameter index missing");
    }

    boost::shared_ptr<const Device> device = getDeviceFromParam();

    dS_Assistant *dsa = dS_Assistant::getInstance();

    uint8_t index = util::str2int(indexs);
    uint16_t value;
    dsa->dS485()->getStatisticValue(device->getDSMdSID(), device->getId(),
                                    index, &value);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    json_object_object_add(res, "index", json_object_new_int(index));
    json_object_object_add(res, "value", json_object_new_int(value));
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdeviceWebHandler::getStatisticValueByName()
{
    std::ostringstream result;
    std::string sname = m_parameters["name"];
    if (sname.empty())
    {
        throw _Exception("Parameter name missing");
    }

    boost::shared_ptr<const Device> device = getDeviceFromParam();

    dS_Assistant *dsa = dS_Assistant::getInstance();
    Database *db = Database::getInstance();
    statistic_field_t stype = STAT_INVALID;

    boost::shared_ptr<std::list<std::pair<statistic_field_t,std::string> > > tnames =
        db->getStatisticTableNames(0);

    std::list<std::pair<statistic_field_t,std::string> >::iterator itname;
    for (itname = tnames->begin();
         itname != tnames->end();
         itname++)
    {
        if (sname.compare((*itname).second) == 0)
        {
            stype = (*itname).first;
            break;
        }
    }
    if (stype == STAT_INVALID)
    {
        throw Exception("Parameter name is invalid");
    }

    uint8_t offset;
    uint8_t length;
    uint16_t value;
    dsa->dS485()->getStatisticValue(device->getDSMdSID(), device->getId(), 0, &value);

    boost::shared_ptr<StatisticDevice> sdata(new StatisticDevice(value & 0xff));
    sdata->getOffsetForType(stype, offset, length);

    for (uint8_t n = offset; n < offset + length; n += 2)
    {
        dsa->dS485()->getStatisticValue(device->getDSMdSID(), device->getId(), n >> 1, &value);
        sdata->setValue(n, value, 0);
    }
    sdata->parseValues();

    boost::shared_ptr<StatisticValue> svalue = sdata->getValueForType(stype);

    db->saveStatisticDevice(0, device->getDSID(), device->getDSMdSID(), device->getId(), stype, svalue->m_value, 0);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    json_object_object_add(res, "name", json_object_new_string(sname.c_str()));
    json_object_object_add(res, "value", json_object_new_int(svalue->m_value));
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdeviceWebHandler::testTransmissionQuality()
{
    std::ostringstream result;

    boost::shared_ptr<const Device> device = getDeviceFromParam();

    dS_Assistant *dsa = dS_Assistant::getInstance();

    uint8_t down;
    uint16_t up;
    dsa->dS485()->testDeviceTransmissionQuality(device->getDSMdSID(),
                                                device->getId(), &down, &up);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    json_object_object_add(res, "upstream", json_object_new_int(up));
    json_object_object_add(res, "downstream", json_object_new_int(down));
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

void JSONdeviceWebHandler::process()
{
    try
    {
        std::string result;
        std::string header = http_ok();

        if (m_command == "blink")
        {
            result = blink();
        }
        else if (m_command == "testTransmissionQuality")
        {
            result = testTransmissionQuality();
        }
        else if (m_command == "getStatisticValueByIndex")
        {
            result = getStatisticValueByIndex();
        }
        else if (m_command == "getStatisticValueByName")
        {
            result = getStatisticValueByName();
        }
        else if (m_command == "resetStatistic")
        {
            result = resetStatistic();
        }
        else
        {
            failure("unknown function called");
            return;
        }

        mg_write(m_connection, header.c_str(), header.length());
        mg_write(m_connection, result.c_str(), result.length());
    }
    catch (Exception ex)
    {
        failure(ex.getMessage().c_str());
    }
    catch (...)
    {
        failure("exception raised while processing call " + m_command);
    }
}

} // namespace

